/*
    Copyright (c) 2010, 2011 aizo ag, Zurich, Switzerland

    Author: Sergey 'Jin' Bostandzhyan <jin@dev.digitalstrom.org>

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifdef HAVE_CONFIG_H
    #include "autoconfig.h"
#endif

#include <autosprintf.h>
#include <digitalSTROM/ds485-client-interface.h>

#include "dsmapi-wrapper.h"
#include "exceptions.h"
#include "traffic-decode.h"
#include "util.h"
#include "log.h"
#include "gettext_helper.h"

#define DEFAULT_SYNC_CALL_TIMEOUT   20 // seconds

/// \file dsmapi-wrapper.h

using gnu::autosprintf;

namespace dSA
{

dSMApiWrapper::dSMApiWrapper(std::string ds485_daemon_url)
{
    int ret;

    boost::mutex::scoped_lock lock(m_libdsm_mutex);

    m_dsm_api_handle = DsmApiInitialize();
    if (m_dsm_api_handle == NULL)
    {
        throw _Exception("could not initialize dSM API library");
    }

    ret = DsmApiOpen(m_dsm_api_handle, ds485_daemon_url.c_str(), 0);
    if (ret < 0)
    {
        throw _Exception("could not open connection to " + ds485_daemon_url +
                         " (" + util::str(ret) + ")");
    }

    // register callbacks
    DsmApiCallback_t cb;
    cb.function = reinterpret_cast<void *>(dSMApiWrapper::linkStateCallback);
    cb.arg = this;
    DsmApiSetLinkStateCallback(m_dsm_api_handle, &cb, NULL);

    cb.function = (void *) dSMApiWrapper::busChangeCallback;
    cb.arg = this;
    DsmApiSetBusChangeCallback(m_dsm_api_handle, &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceActiveCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACCESSIBILITY,
                           EVENT_DEVICE_ACCESSIBILITY_ON,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceInactiveCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACCESSIBILITY,
                           EVENT_DEVICE_ACCESSIBILITY_OFF,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::devicePingCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_TEST_DEVICE_PRESENCE,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceActionCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_ACTION,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceLocalActionCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_LOCAL_ACTION,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceConfigCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_CONFIG,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::commUpstreamShortCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_UPSTREAM_SHORT,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::commUpstreamLongCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_UPSTREAM_LONG,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::commDownstreamShortCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_SHORT,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::commDownstreamLong0Callback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_NO_PARAMETER,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::commDownstreamLong8Callback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_8BIT_PARAMETER,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::commDownstreamLong16Callback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_COMMUNICATION_LOG,
                           EVENT_COMMUNICATION_LOG_DOWNSTREAM_LONG_16BIT_PARAMETER,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::dsmStatusCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DSM_STATE_CHANGE,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::dsmErrorCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DSM_ERROR,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::dsmRebootRequestCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_REQUEST,
                           DSM_REBOOT,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceStatisticsCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_DEVICE_STATISTICS,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceTestTransmissionQualityCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_EVENT,
                           EVENT_TEST_TRANSMISSION_QUALITY,
                           0,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceRemovedByDsuidCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_REQUEST,
                           CIRCUIT_REMOVE_DEVICE,
                           CIRCUIT_REMOVE_DEVICE_BY_DSUID,
                           &cb, NULL);

    cb.function = (void *) dSMApiWrapper::deviceRemovedByIdCallback;
    cb.arg = this;
    DsmApiSetCallback(m_dsm_api_handle, DS485_CONTAINER_REQUEST,
                           CIRCUIT_REMOVE_DEVICE,
                           CIRCUIT_REMOVE_DEVICE_BY_ID,
                           &cb, NULL);
}

void dSMApiWrapper::addListener(dSMApiListener* listener)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    m_callback_listeners.push_back(listener);
}

void dSMApiWrapper::removeListener(dSMApiListener* listener)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == listener)
        {
            i = m_callback_listeners.erase(i);
            break;
        }

        i++;
    }
}

void dSMApiWrapper::linkStateCallback(void* _arg, bool _link)
{
    static_cast<dSMApiWrapper *>(_arg)->notifyLinkState(_link);
}

void dSMApiWrapper::notifyLinkState(bool _link)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->linkStateCallback(_link);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::busChangeCallback(void* arg, dsuid_t *id, int flag)
{
    static_cast<dSMApiWrapper *>(arg)->notifyBusChange(id, flag);
}

void dSMApiWrapper::notifyBusChange(dsuid_t *id, int flag)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->busChangeCallback(id, flag);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceActiveCallback(uint8_t error_code, void *arg,
                                         dsuid_t src_dsm_dsuid,
                                         dsuid_t dest_dsm_dsuid,
                                         uint16_t device_id,
                                         uint16_t zone_id,
                                         uint16_t vendor_id,
                                         uint16_t product_id,
                                         uint16_t function_id,
                                         uint16_t version,
                                         dsuid_t dsuid)
{
    (void)product_id;
    (void)function_id;
    (void)version;
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceActive(error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id, zone_id, vendor_id,
            dsuid);
}

void dSMApiWrapper::notifyDeviceActive(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                       dsuid_t dest_dsm_dsuid, uint16_t device_id,
                                       uint16_t zone_id, uint16_t vendor_id,
                                       dsuid_t dsuid)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener *>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceActiveCallback(error_code, src_dsm_dsuid,
                                           dest_dsm_dsuid, device_id,
                                           zone_id, vendor_id, dsuid);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }

            i++;
        }
    }
}

void dSMApiWrapper::deviceInactiveCallback(uint8_t error_code, void *arg,
                                           dsuid_t src_dsm_dsuid,
                                           dsuid_t dest_dsm_dsuid,
                                           uint16_t device_id,
                                           uint16_t zone_id,
                                           uint16_t vendor_id,
                                           uint16_t product_id,
                                           uint16_t function_id,
                                           uint16_t version,
                                           dsuid_t dsuid)
{
    (void)product_id;
    (void)function_id;
    (void)version;
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceInactive(error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id, zone_id, vendor_id,
            dsuid);
}

void dSMApiWrapper::notifyDeviceInactive(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                         dsuid_t dest_dsm_dsuid,
                                         uint16_t device_id, uint16_t zone_id,
                                         uint16_t vendor_id,
                                         dsuid_t dsuid)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceInactiveCallback(error_code, src_dsm_dsuid,
                                             dest_dsm_dsuid, device_id, zone_id,
                                             vendor_id, dsuid);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}


void dSMApiWrapper::devicePingCallback(uint8_t error_code, void *arg,
                                         dsuid_t src_dsm_dsuid,
                                         dsuid_t dest_dsm_dsuid,
                                         uint16_t device_id,
                                         uint8_t present)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDevicePing(error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id, present);
}

void dSMApiWrapper::notifyDevicePing(uint8_t error_code, dsuid_t src_dsm_dsuid,
                                        dsuid_t dest_dsm_dsuid,
                                        uint16_t device_id, uint8_t present)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->devicePingCallback(error_code, src_dsm_dsuid,
                                         dest_dsm_dsuid, device_id, present);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceStatisticsCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceStatistics(error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id, offset, value);
}

void dSMApiWrapper::notifyDeviceStatistics(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t offset, uint16_t value)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceStatisticsCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id, offset, value);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceActionCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceActionCallback(error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id,
            zone_id, group_id, button_nr, click_type);
}
void dSMApiWrapper::notifyDeviceActionCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
        uint8_t click_type)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceActionCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id,
                        zone_id, group_id, button_nr, click_type);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceLocalActionCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceLocalActionCallback(
            error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id,
            zone_id, state);
}
void dSMApiWrapper::notifyDeviceLocalActionCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint16_t zone_id, uint8_t state)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceLocalActionCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id,
                        zone_id, state);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceConfigCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceConfigCallback(error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id,
            config_class, config_index, value);
}

void dSMApiWrapper::notifyDeviceConfigCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceConfigCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id,
                        config_class, config_index, value);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceTestTransmissionQualityCallback(uint8_t error_code,
        void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    static_cast<dSMApiWrapper *>(arg)->
        notifyDeviceTestTransmissionQualityCallback(error_code,
                src_dsm_dsuid, dest_dsm_dsuid, device_id,
                downstream_quality, upstream_quality);
}

void dSMApiWrapper::notifyDeviceTestTransmissionQualityCallback(
        uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t downstream_quality, uint16_t upstream_quality)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceTestTransmissionQualityCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id,
                        downstream_quality, upstream_quality);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceRemovedByDsuidCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceRemovedByDsuidCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, dsuid);
}

void dSMApiWrapper::notifyDeviceRemovedByDsuidCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceRemovedByDsuidCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, dsuid);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::deviceRemovedByIdCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDeviceRemovedByIdCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, device_id);
}

void dSMApiWrapper::notifyDeviceRemovedByIdCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->deviceRemovedByIdCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commUpstreamShortCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags, uint8_t cflags)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommUpstreamShortCallback(
            error_code,
            src_dsm_dsuid, dest_dsm_dsuid, device_id,
            circuit_id, is_resend, is_sensor, key_type, click_type, quality,
            flags, cflags);
}

void dSMApiWrapper::notifyCommUpstreamShortCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
        uint8_t key_type, uint8_t click_type,
        uint8_t quality, uint8_t flags, uint8_t cflags)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commUpstreamShortCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id,
                        circuit_id, is_resend, is_sensor, key_type, click_type,
                        quality, flags, cflags);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commUpstreamLongCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommUpstreamLongCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, device_id,
            circuit_id, mainqualifier, subqualifier, data, quality,
            flags, cflags);
}

void dSMApiWrapper::notifyCommUpstreamLongCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
        uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
        uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commUpstreamLongCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, device_id,
                        circuit_id, mainqualifier, subqualifier, data,
                        quality, flags, cflags);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamShortCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamShortCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, command);
}

void dSMApiWrapper::notifyCommDownstreamShortCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamShortCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, command);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong0Callback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong0Callback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, command);
}

void dSMApiWrapper::notifyCommDownstreamLong0Callback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamLong0Callback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, command);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong8Callback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
        uint8_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong8Callback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, command, data);
}

void dSMApiWrapper::notifyCommDownstreamLong8Callback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
        uint8_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamLong8Callback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, command, data);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::commDownstreamLong16Callback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
        uint16_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyCommDownstreamLong16Callback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid, command, data);
}

void dSMApiWrapper::notifyCommDownstreamLong16Callback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
        uint16_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->commDownstreamLong16Callback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid, command, data);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::dsmStatusCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint8_t hardware, uint8_t communicationdsp, uint8_t temperature,
        uint8_t configuration, uint8_t ds485bus, uint8_t powerline)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDsmStatusCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid,
            hardware, communicationdsp, temperature,
            configuration, ds485bus, powerline);
}

void dSMApiWrapper::notifyDsmStatusCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint8_t hardware, uint8_t communicationdsp, uint8_t temperature,
        uint8_t configuration, uint8_t ds485bus, uint8_t powerline)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->dsmStatusCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid,
                        hardware, communicationdsp, temperature,
                        configuration, ds485bus, powerline);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::dsmErrorCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint16_t error, uint32_t data)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDsmErrorCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid,
            error, data);
}

void dSMApiWrapper::notifyDsmErrorCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint16_t error, uint32_t data)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->dsmErrorCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid,
                        error, data);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}

void dSMApiWrapper::dsmRebootRequestCallback(uint8_t error_code, void *arg,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid)
{
    static_cast<dSMApiWrapper *>(arg)->notifyDsmRebootRequestCallback(
            error_code, src_dsm_dsuid, dest_dsm_dsuid);
}

void dSMApiWrapper::notifyDsmRebootRequestCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_listener_mutex);
    std::list<dSMApiListener*>::iterator i = m_callback_listeners.begin();

    while (i != m_callback_listeners.end())
    {
        if ((*i) == NULL)
        {
            i = m_callback_listeners.erase(i);
        }
        else
        {
            try
            {
                (*i)->dsmRebootRequestCallback(error_code,
                        src_dsm_dsuid, dest_dsm_dsuid);
            }
            catch (Exception &ex)
            {
                FILE_LOG(logERROR) << ex.getMessage();
            }
            i++;
        }
    }
}


void dSMApiWrapper::getOwnDSUID(dsuid_t* dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetOwnDSUID(m_dsm_api_handle, dsuid);
    checkRetcode(ret);
}

bool dSMApiWrapper::getLinkState()
{
    bool state;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetLinkState(m_dsm_api_handle, &state);
    checkRetcode(ret);
    return state;
}

void dSMApiWrapper::restartRegistration(dsuid_t dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_restart_registration(m_dsm_api_handle, dsm_dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::restartMeter(dsuid_t dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMReboot(m_dsm_api_handle, dsm_dsuid);
    checkRetcode(ret);
    DSUID_LOG(logINFO, dsm_dsuid) << "Reboot dSM";
}

int dSMApiWrapper::getBusMembersPerType(dsuid_t *device_list, uint16_t types)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DsmApiGetBusMembersPerType(m_dsm_api_handle, device_list,
                                         DSM_API_MAX_DEVICES, types);
    if (ret < 0)
    {
        throw _ExceptionWithCode("could not get bus members", ret);
    }

    return ret;
}

int dSMApiWrapper::getBusMemberType(dsuid_t busMember)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint8_t type;

    int ret = BusMember_get_type(m_dsm_api_handle, busMember, &type);
    checkRetcode(ret);
    return (int)type;
}

void dSMApiWrapper::dSMInfo(dsuid_t dsm_dsuid, uint32_t *hw_version,
                           uint32_t *arm_sw_version, uint32_t *dsp_sw_version,
                           uint16_t *api_version, dsid_t *dsid, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    /// \todo make sure to always check the handle and connection state
    ret = ::dSMInfo(m_dsm_api_handle, dsm_dsuid, hw_version, arm_sw_version,
                    dsp_sw_version, api_version, dsid, name);
    checkRetcode(ret);
}

void dSMApiWrapper::dSFInfo(dsuid_t dsm_dsuid, uint32_t *hw_version,
                     uint32_t *arm_sw_version, uint32_t *dsp_sw_version, uint32_t *booster_arm_sw_version,
                     uint16_t *api_version, dsid_t *dsid, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ::dSFInfo(m_dsm_api_handle, dsm_dsuid, hw_version, arm_sw_version, dsp_sw_version,
                    booster_arm_sw_version, api_version, dsid, name);
    checkRetcode(ret);
}

void dSMApiWrapper::dSMState(dsuid_t dsm_dsuid, uint8_t *hardware,
                            uint8_t *communicationdsp, uint8_t *temperature,
                            uint8_t *configuration, uint8_t *ds485bus,
                            uint8_t *powerline)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ::dSMState(m_dsm_api_handle, dsm_dsuid, hardware, communicationdsp,
                     temperature, configuration, ds485bus, powerline);
    checkRetcode(ret);
}

void dSMApiWrapper::getPropertyFlags(dsuid_t dsm_dsuid, uint8_t *flags)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_flags(m_dsm_api_handle, dsm_dsuid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::setPropertyFlags(dsuid_t dsm_dsuid, uint8_t flags)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_flags(m_dsm_api_handle, dsm_dsuid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::getPropertyTemperatureLimit(dsuid_t dsm_dsuid,
                                                uint8_t *warnlevel,
                                                uint8_t *errorlevel)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_temperature_limits(m_dsm_api_handle, dsm_dsuid,
                                               warnlevel, errorlevel);
    checkRetcode(ret);
}

void dSMApiWrapper::setPropertyTemperatureLimit(dsuid_t dsm_dsuid,
                                                uint8_t warnlevel,
                                                uint8_t errorlevel)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_temperature_limits(m_dsm_api_handle, dsm_dsuid,
                                               warnlevel, errorlevel);
    checkRetcode(ret);
}

void dSMApiWrapper::getDSMtemperature(dsuid_t dsm_dsuid, uint16_t *temperature)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_get_current_temperature(m_dsm_api_handle, dsm_dsuid,
                                                    temperature);
    checkRetcode(ret);
}

void dSMApiWrapper::getDSMMaxTemperature(dsuid_t dsm_dsuid, uint16_t *temperature)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_get_max_temperature(m_dsm_api_handle, dsm_dsuid,
                                                temperature);
    checkRetcode(ret);
}

void dSMApiWrapper::resetDSMMaxTemperature(dsuid_t dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_reset_max_temperature(m_dsm_api_handle, dsm_dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::getApartmentId(dsuid_t dsm_dsuid, uint8_t *id)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_apartment_id(m_dsm_api_handle, dsm_dsuid, id);
    checkRetcode(ret);
}

void dSMApiWrapper::setApartmentId(dsuid_t dsm_dsuid, uint8_t id)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_apartment_id(m_dsm_api_handle, dsm_dsuid, id);
    checkRetcode(ret);
}

void dSMApiWrapper::getUpstreamFrequency(dsuid_t dsm_dsuid, uint8_t *freq)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_get_upstream_frequency(m_dsm_api_handle, dsm_dsuid,
                                               freq);
    checkRetcode(ret);
}

void dSMApiWrapper::setUpstreamFrequency(dsuid_t dsm_dsuid, uint8_t freq)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMProperties_set_upstream_frequency(m_dsm_api_handle, dsm_dsuid,
                                               freq);
    checkRetcode(ret);
}

void dSMApiWrapper::getStatisticUpstream(dsuid_t dsm_dsuid, uint8_t index,
        uint8_t *length, uint8_t *statistics)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMStatistics_get_upstream(m_dsm_api_handle, dsm_dsuid, index,
            length, statistics);
    checkRetcode(ret);
}

void dSMApiWrapper::resetStatisticUpstream(dsuid_t dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMStatistics_reset_upstream(m_dsm_api_handle, dsm_dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::getdSFStatisticsDSP(dsuid_t dsf_dsuid, uint8_t dsp_nr,
        uint8_t index, uint8_t *length, uint8_t *statistics)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSFStatistics_get_DSP(m_dsm_api_handle, dsf_dsuid, dsp_nr, index, length, statistics);
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationId(dsuid_t dsm_dsuid, uint32_t *hash,
        uint32_t *configcount)
{
    int ret;
    uint32_t eventCount;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = dSMConfig_get_hash(m_dsm_api_handle, dsm_dsuid, hash, configcount, &eventCount);
    checkRetcode(ret);
}


void dSMApiWrapper::getZoneCount(dsuid_t dsm_dsuid, uint8_t *count)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneCount(m_dsm_api_handle, dsm_dsuid, count);
    checkRetcode(ret);
}

void dSMApiWrapper::getZoneInfo(dsuid_t dsm_dsuid, uint8_t zone_index,
                               uint16_t *zone_id, uint8_t *virtual_zone_id,
                               uint8_t *number_of_groups, uint8_t *name)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneInfo_by_index(m_dsm_api_handle, dsm_dsuid, zone_index, zone_id,
                            virtual_zone_id, number_of_groups, name);
    checkRetcode(ret);
}

void dSMApiWrapper::getActiveDevicesCountInZone(dsuid_t dsm_dsuid,
                                               uint16_t zone_id,
                                               uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_only_active(m_dsm_api_handle, dsm_dsuid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getInactiveDevicesCountInZone(dsuid_t dsm_dsuid,
                                                  uint16_t zone_id,
                                                  uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_only_inactive(m_dsm_api_handle, dsm_dsuid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getDevicesCountInZone(dsuid_t dsm_dsuid,
                                          uint16_t zone_id,
                                          uint16_t *number_of_devices)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = ZoneDeviceCount_all(m_dsm_api_handle, dsm_dsuid, zone_id,
                                      number_of_devices);
    checkRetcode(ret);
}

void dSMApiWrapper::getActiveDeviceInfoInZone(dsuid_t dsm_dsuid,
                                             uint16_t req_zone_id,
                                             uint16_t device_index,
                                             uint16_t *device_id,
                                             uint16_t *vendor_id,
                                             uint16_t *product_id,
                                             uint16_t *function_id,
                                             uint16_t *version,
                                             uint16_t *zone_id,
                                             uint8_t *state,
                                             uint8_t *locked,
                                             uint8_t *output_mode,
                                             uint8_t *groups,
                                             uint8_t *name,
                                             dsuid_t *dsuid,
                                             uint8_t *lt_mode)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_index_only_active(m_dsm_api_handle, dsm_dsuid,
                                          req_zone_id, device_index, device_id,
                                          vendor_id, product_id, function_id,
                                          version, zone_id, state, locked,
                                          output_mode, lt_mode, groups,
                                          name, dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::getDeviceInfoById(dsuid_t dsm_dsuid, uint16_t req_device_id,
                                 uint16_t *device_id, uint16_t *vendor_id,
                                 uint16_t *product_id, uint16_t *function_id,
                                 uint16_t *version, uint16_t *zone_id,
                                 uint8_t *state, uint8_t *locked,
                                 uint8_t *output_mode, uint8_t *groups,
                                 uint8_t *name, dsuid_t *dsuid,
                                 uint8_t *lt_mode)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_device_id(m_dsm_api_handle, dsm_dsuid, req_device_id,
                                  device_id, vendor_id, product_id,
                                  function_id, version, zone_id, state,
                                  locked, output_mode, lt_mode, groups,
                                  name, dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::getDeviceInfoByIndex(dsuid_t dsm_dsuid, uint16_t req_zone_id, uint16_t req_index,
                                 uint16_t *device_id, uint16_t *vendor_id,
                                 uint16_t *product_id, uint16_t *function_id,
                                 uint16_t *version, uint16_t *zone_id,
                                 uint8_t *state, uint8_t *locked, uint8_t *output_mode,
                                 uint8_t *groups, uint8_t *name,
                                 dsuid_t *dsuid, uint8_t *lt_mode)
{
    int ret;
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    ret = DeviceInfo_by_index(m_dsm_api_handle, dsm_dsuid, req_zone_id,
                                  req_index,
                                  device_id, vendor_id, product_id,
                                  function_id, version, zone_id, state,
                                  locked, output_mode, lt_mode, groups,
                                  name, dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::blinkDevice(dsuid_t dsm_dsuid, uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceActionRequest_action_blink(m_dsm_api_handle, dsm_dsuid,
                                               device_id);
    checkRetcode(ret);
}

void dSMApiWrapper::setConfigurationValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t value)
{

    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = 0;
#if DSM_API_VERSION > 0x300
    ret = DeviceConfig_set_sync(m_dsm_api_handle, dsm_dsuid,
            device_id, config_class, config_index, value,
            DEFAULT_SYNC_CALL_TIMEOUT);

#else
    ret = DeviceConfig_set(m_dsm_api_handle, dsm_dsuid,
            device_id, config_class, config_index, value);
    sleep(DEFAULT_SYNC_CALL_TIMEOUT);
#endif
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint8_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceConfig_get_sync_8(m_dsm_api_handle, dsm_dsuid,
            device_id, config_class, config_index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getConfigurationValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t config_class, uint8_t config_index, uint16_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceConfig_get_sync_16(m_dsm_api_handle, dsm_dsuid,
            device_id, config_class, config_index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t offset, uint8_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint16_t data;
    int ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsuid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            &data);
    checkRetcode(ret);
    if (offset & 1)
        *value = data >> 8;
    else
        *value = data & 0xff;
}

void dSMApiWrapper::getStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t offset, uint16_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsuid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::getStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t offset, uint32_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint16_t datal, datah;
    int ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsuid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            &datal);
    checkRetcode(ret);
    ret = DeviceStatisticsRequest_get_sync(m_dsm_api_handle, dsm_dsuid,
            device_id, offset,
            DEFAULT_SYNC_CALL_TIMEOUT,
            &datah);
    checkRetcode(ret);
    *value = datah << 16 | datal;
}

void dSMApiWrapper::reqStatisticValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t offset)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceStatisticsRequest_get(m_dsm_api_handle, dsm_dsuid,
            device_id, offset);
    checkRetcode(ret);
}

void dSMApiWrapper::getSensorValue(dsuid_t dsm_dsuid, uint16_t device_id,
        uint8_t index, uint16_t *value)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = DeviceSensor_get_value_sync(m_dsm_api_handle, dsm_dsuid,
            device_id, index,
            DEFAULT_SYNC_CALL_TIMEOUT,
            value);
    checkRetcode(ret);
}

void dSMApiWrapper::testDeviceTransmissionQuality(dsuid_t dsm_dsuid,
                                                  uint16_t device_id,
                                                  uint8_t *downstreamQ,
                                                  uint16_t *upstreamQ)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = TestTransmissionQuality_get_sync(m_dsm_api_handle, dsm_dsuid,
                                               device_id,
                                               DEFAULT_SYNC_CALL_TIMEOUT,
                                               downstreamQ,
                                               upstreamQ);
    checkRetcode(ret);
}

void dSMApiWrapper::reqDeviceTransmissionQuality(dsuid_t dsm_dsuid,
                                                 uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = TestTransmissionQuality_get (m_dsm_api_handle, dsm_dsuid, device_id);
    checkRetcode(ret);
}

void dSMApiWrapper::reqDevicePresenceTest(dsuid_t dsm_dsuid,
                                          uint16_t device_id)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
#if DSM_API_VERSION >= 0x107
    int ret = TestDevicePresence(m_dsm_api_handle, dsm_dsuid, device_id);
    checkRetcode(ret);
#endif
}

void dSMApiWrapper::setDisplayMode(dsuid_t dsm_dsuid, uint8_t flags)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_set_MMI_flag(m_dsm_api_handle, dsm_dsuid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::getDisplayMode(dsuid_t dsm_dsuid, uint8_t* flags)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_get_MMI_flag(m_dsm_api_handle, dsm_dsuid, flags);
    checkRetcode(ret);
}

void dSMApiWrapper::setCommunicationLogging(dsuid_t dsm_dsuid, uint8_t state)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_set_communication_log_state(m_dsm_api_handle,
                                                       dsm_dsuid, state);
    checkRetcode(ret);
}

void dSMApiWrapper::getCommunicationLogging(dsuid_t dsm_dsuid, uint8_t* state)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_get_communication_log_state(m_dsm_api_handle,
                                                       dsm_dsuid, state);
    checkRetcode(ret);
}

void dSMApiWrapper::getDspConfig(dsuid_t dsm_dsuid, DsmDspConfigData_t* config)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint8_t buffer[96];
    uint8_t size;
    int ret = dSMConfig_get_DSP_config(m_dsm_api_handle, dsm_dsuid, &size,
                                       buffer);
    checkRetcode(ret);

    memcpy(config, buffer, std::min(static_cast<std::size_t>(size),
                                    sizeof (DsmDspConfigData_t)));
}

void dSMApiWrapper::writeDspConfig(dsuid_t dsm_dsuid, const DsmDspConfigData_t &config)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    uint8_t buffer[96];
    memcpy(buffer, &config, sizeof (DsmDspConfigData_t));
    int ret = dSMConfig_set_DSP_config(m_dsm_api_handle, dsm_dsuid,
                                       sizeof (DsmDspConfigData_t), buffer);
    checkRetcode(ret);
}

void dSMApiWrapper::restartDsp(dsuid_t dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMDebugMode_restart_dsp(m_dsm_api_handle, dsm_dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::pingAutomatic(dsuid_t dsm_dsuid, uint16_t auto_interval_time)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_automatic(m_dsm_api_handle, dsm_dsuid,
                                           auto_interval_time);
    checkRetcode(ret);
}

void dSMApiWrapper::pingOneTimeAutomatic(dsuid_t dsm_dsuid,
                            uint16_t interval_time, uint16_t auto_interval_time)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_one_time_automatic(m_dsm_api_handle, dsm_dsuid,
                                            interval_time, auto_interval_time);
    checkRetcode(ret);
}

void dSMApiWrapper::pingOneTimeOff(dsuid_t dsm_dsuid, uint16_t interval_time)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_one_time_off(m_dsm_api_handle, dsm_dsuid,
                                                                interval_time);
    checkRetcode(ret);
}

void dSMApiWrapper::pingOff(dsuid_t dsm_dsuid)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = CircuitPingDevices_off(m_dsm_api_handle, dsm_dsuid);
    checkRetcode(ret);
}

void dSMApiWrapper::getPingIntervals(dsuid_t dsm_dsuid, uint16_t *interval_auto,
        uint16_t *interval_device)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_get_ping_intervals(m_dsm_api_handle, dsm_dsuid,
                                                interval_auto, interval_device);
    checkRetcode(ret);
}

void dSMApiWrapper::setPingIntervals(dsuid_t dsm_dsuid, uint16_t interval_auto,
                                                    uint16_t interval_device)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = dSMProperties_set_ping_intervals(m_dsm_api_handle, dsm_dsuid,
                                            interval_auto, interval_device);
    checkRetcode(ret);
}


void dSMApiWrapper::protobufMessageRequest(dsuid_t dsm_dsuid,
                                           uint16_t request_size,
                                           uint8_t *request,
                                           uint16_t *response_size,
                                           uint8_t *response)
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    int ret = UserProtobufMessageRequest(m_dsm_api_handle, dsm_dsuid,
                                         request_size, request,
                                         response_size, response);
    checkRetcode(ret);
}

void dSMApiWrapper::checkRetcode(int ret)
{
    if (ret != 0)
    {
        throw _ExceptionWithCode(std::string(ds485c_strerror(ret)) + ": " +
                                 util::str(ret), ret);
    }
}

dSMApiWrapper::~dSMApiWrapper()
{
    boost::mutex::scoped_lock lock(m_libdsm_mutex);
    if (m_dsm_api_handle)
    {
        DsmApiClose(m_dsm_api_handle);
        DsmApiCleanup(m_dsm_api_handle);
    }
}

dSMApiSelectiveListener::CBData::CBData(
        dSMApiSelectiveListener::cb_selection type) : m_data_type(type)
{
    m_timestamp = time(NULL);
}

dSMApiSelectiveListener::CBData::CBData() :
        m_data_type(CB_MAX)
{
    m_timestamp = time(NULL);
}

dSMApiSelectiveListener::cb_selection
    dSMApiSelectiveListener::CBData::getType() const
{
    return m_data_type;
}

time_t dSMApiSelectiveListener::CBData::getTimestamp() const
{
    return m_timestamp;
}

dSMApiSelectiveListener::CBCommonData::CBCommonData(
    dSMApiSelectiveListener::cb_selection type, uint8_t error_code,
    dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid) :
    CBData(type), m_error_code(error_code), m_src_dsm_dsuid(src_dsm_dsuid), m_dest_dsm_dsuid(dest_dsm_dsuid)
{}

uint8_t dSMApiSelectiveListener::CBCommonData::getErrorCode() const
{
    return m_error_code;
}

dsuid_t dSMApiSelectiveListener::CBCommonData::getDSMSourceDSUID() const
{
    return m_src_dsm_dsuid;
}

dsuid_t dSMApiSelectiveListener::CBCommonData::getDSMDestinationDSUID() const
{
    return m_dest_dsm_dsuid;
}

dSMApiSelectiveListener::BasicDeviceData::BasicDeviceData(
    dSMApiSelectiveListener::cb_selection type,
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id) :
    CBCommonData(type, error_code, src_dsm_dsuid, dest_dsm_dsuid),
    m_device_id(device_id)
{}

uint16_t dSMApiSelectiveListener::BasicDeviceData::getDeviceID() const
{
    return m_device_id;
}

dSMApiSelectiveListener::DeviceData::DeviceData(
    dSMApiSelectiveListener::cb_selection type,
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint16_t zone_id, uint16_t vendor_id, dsuid_t dsuid) :
        BasicDeviceData(type, error_code, src_dsm_dsuid, dest_dsm_dsuid,
        device_id), m_zone_id(zone_id), m_vendor_id(vendor_id)
{
    (void)dsuid;
}

uint16_t dSMApiSelectiveListener::DeviceData::getZoneID() const
{
    return m_zone_id;
}

uint16_t dSMApiSelectiveListener::DeviceData::getVendorID() const
{
    return m_vendor_id;
}

dSMApiSelectiveListener::DeviceActiveData::DeviceActiveData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint16_t zone_id, uint16_t vendor_id,
    dsuid_t dsuid) :
        DeviceData(dSMApiSelectiveListener::DsmApiDeviceActive, error_code,
        src_dsm_dsuid, dest_dsm_dsuid, device_id, zone_id, vendor_id,
        dsuid)
{}

dSMApiSelectiveListener::DeviceInactiveData::DeviceInactiveData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint16_t zone_id, uint16_t vendor_id,
    dsuid_t dsuid) :
        DeviceData(dSMApiSelectiveListener::DsmApiDeviceInactive, error_code,
        src_dsm_dsuid, dest_dsm_dsuid, device_id, zone_id, vendor_id,
        dsuid)
{}

dSMApiSelectiveListener::DeviceTestQualityData::DeviceTestQualityData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint8_t downstream, uint16_t upstream) :
        BasicDeviceData(dSMApiSelectiveListener::DsmApiDeviceTestTransmissionQuality,
                error_code, src_dsm_dsuid, dest_dsm_dsuid, device_id),
        m_down(downstream), m_up(upstream)
{}

uint8_t dSMApiSelectiveListener::DeviceTestQualityData::getDownstream() const
{
    return m_down;
}

uint16_t dSMApiSelectiveListener::DeviceTestQualityData::getUpstream() const
{
    return m_up;
}

dSMApiSelectiveListener::DevicePingData::DevicePingData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint8_t present) :
        BasicDeviceData(dSMApiSelectiveListener::DsmApiDevicePing,
                error_code, src_dsm_dsuid, dest_dsm_dsuid, device_id),
        m_present(present)
{}

uint8_t dSMApiSelectiveListener::DevicePingData::getPresent() const
{
    return m_present;
}

dSMApiSelectiveListener::DeviceStatisticData::DeviceStatisticData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint8_t offset, uint16_t value) :
        BasicDeviceData(dSMApiSelectiveListener::DsmApiDeviceStatistics,
                error_code, src_dsm_dsuid, dest_dsm_dsuid, device_id),
        m_offset(offset), m_value(value)
{}

uint8_t dSMApiSelectiveListener::DeviceStatisticData::getOffset() const
{
    return m_offset;
}

uint16_t dSMApiSelectiveListener::DeviceStatisticData::getValue() const
{
    return m_value;
}

dSMApiSelectiveListener::CommData::CommData(cb_selection type,
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint8_t circuit_id, uint8_t quality,
    uint8_t flags, uint8_t cflags) :
    BasicDeviceData(type, error_code, src_dsm_dsuid, dest_dsm_dsuid, device_id),
    m_circuit_id(circuit_id), m_quality(quality),
    m_flags(flags), m_cflags(cflags)
{}

uint8_t dSMApiSelectiveListener::CommData::getCircuitID() const
{
    return m_circuit_id;
}

uint8_t dSMApiSelectiveListener::CommData::getQuality() const
{
    return m_quality;
}

uint8_t dSMApiSelectiveListener::CommData::getFlags() const
{
    return m_flags;
}

uint8_t dSMApiSelectiveListener::CommData::getCrosstalkFlags() const
{
    return m_cflags;
}

dSMApiSelectiveListener::CommUpstreamShortData::CommUpstreamShortData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint8_t circuit_id, uint8_t is_resend,
    uint8_t is_sensor, uint8_t key_type, uint8_t click_type, uint8_t quality,
    uint8_t flags, uint8_t cflags) :
        CommData(dSMApiSelectiveListener::DsmApiCommUpstreamShort, error_code,
        src_dsm_dsuid, dest_dsm_dsuid, device_id, circuit_id, quality,
        flags, cflags),
        m_is_resend(is_resend), m_is_sensor(is_sensor),
        m_key_type(key_type), m_click_type(click_type)
{}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::isResend() const
{
    return m_is_resend;
}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::isSensor() const
{
    return m_is_sensor;
}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::getKeyType() const
{
    return m_key_type;
}

uint8_t dSMApiSelectiveListener::CommUpstreamShortData::getClickType() const
{
    return m_click_type;
}

dSMApiSelectiveListener::CommUpstreamLongData::CommUpstreamLongData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t device_id, uint8_t circuit_id, uint8_t mainqualifier,
    uint8_t subqualifier, uint32_t data, uint8_t quality,
    uint8_t flags, uint8_t cflags) :
        CommData(dSMApiSelectiveListener::DsmApiCommUpstreamLong, error_code,
        src_dsm_dsuid, dest_dsm_dsuid, device_id, circuit_id, quality,
        flags, cflags),
        m_mainqualifier(mainqualifier), m_subqualifier(subqualifier),
        m_data(data)
{}

uint8_t dSMApiSelectiveListener::CommUpstreamLongData::getMainQualifier() const
{
    return m_mainqualifier;
}

uint8_t dSMApiSelectiveListener::CommUpstreamLongData::getSubQualifier() const
{
    return m_subqualifier;
}

uint32_t dSMApiSelectiveListener::CommUpstreamLongData::getData() const
{
    return m_data;
}

dSMApiSelectiveListener::CommDownstreamData::CommDownstreamData(
    dSMApiSelectiveListener::cb_selection type, uint8_t error_code,
    dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command) :
    CBCommonData(type, error_code, src_dsm_dsuid, dest_dsm_dsuid),
    m_command(command)
{}

uint8_t dSMApiSelectiveListener::CommDownstreamData::getCommand() const
{
    return m_command;
}

dSMApiSelectiveListener::CommDownstreamShortData::CommDownstreamShortData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint8_t command) :
        CommDownstreamData(dSMApiSelectiveListener::DsmApiCommDownstreamShort,
        error_code, src_dsm_dsuid, dest_dsm_dsuid, command)
{}

dSMApiSelectiveListener::CommDownstreamLong0Data::CommDownstreamLong0Data(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint8_t command) :
        CommDownstreamData(dSMApiSelectiveListener::DsmApiCommDownstreamLong0,
        error_code, src_dsm_dsuid, dest_dsm_dsuid, command)
{}

dSMApiSelectiveListener::CommDownstreamLong8Data::CommDownstreamLong8Data(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint8_t command, uint8_t data) :
        CommDownstreamData(dSMApiSelectiveListener::DsmApiCommDownstreamLong8,
        error_code, src_dsm_dsuid, dest_dsm_dsuid, command), m_data(data)
{}

uint8_t dSMApiSelectiveListener::CommDownstreamLong8Data::getData() const
{
    return m_data;
}

dSMApiSelectiveListener::CommDownstreamLong16Data::CommDownstreamLong16Data(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint8_t command, uint16_t data) :
        CommDownstreamData(dSMApiSelectiveListener::DsmApiCommDownstreamLong16,
        error_code, src_dsm_dsuid, dest_dsm_dsuid, command), m_data(data)
{}

uint16_t dSMApiSelectiveListener::CommDownstreamLong16Data::getData() const
{
    return m_data;
}

dSMApiSelectiveListener::DsmStatusData::DsmStatusData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint8_t hardware, uint8_t communicationdsp,
    uint8_t temperature, uint8_t configuration,
    uint8_t ds485bus, uint8_t powerline) :
        CBCommonData(dSMApiSelectiveListener::DsmApiDsmState,
                error_code, src_dsm_dsuid, dest_dsm_dsuid),
        m_hardware(hardware), m_communicationdsp(communicationdsp),
        m_temperature(temperature), m_configuration(configuration),
        m_ds485bus(ds485bus), m_powerline(powerline)
{}

uint8_t dSMApiSelectiveListener::DsmStatusData::getHardware() const
{
    return m_hardware;
}

uint8_t dSMApiSelectiveListener::DsmStatusData::getCommunicationDSP() const
{
    return m_communicationdsp;
}

uint8_t dSMApiSelectiveListener::DsmStatusData::getTemperature() const
{
    return m_temperature;
}

uint8_t dSMApiSelectiveListener::DsmStatusData::getConfiguration() const
{
    return m_configuration;
}

uint8_t dSMApiSelectiveListener::DsmStatusData::getDS485Bus() const
{
    return m_ds485bus;
}

uint8_t dSMApiSelectiveListener::DsmStatusData::getPowerline() const
{
    return m_powerline;
}

dSMApiSelectiveListener::DsmErrorData::DsmErrorData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
    uint16_t error, uint32_t data) :
        CBCommonData(dSMApiSelectiveListener::DsmApiDsmError,
                error_code, src_dsm_dsuid, dest_dsm_dsuid),
        m_error(error), m_data(data)
{
}

uint16_t dSMApiSelectiveListener::DsmErrorData::getError() const
{
    return m_error;
}

uint32_t dSMApiSelectiveListener::DsmErrorData::getData() const
{
    return m_data;
}

dSMApiSelectiveListener::DsmRebootRequestData::DsmRebootRequestData(
    uint8_t error_code, dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid) :
        CBCommonData(dSMApiSelectiveListener::DsmApiDsmRebootRequest,
                error_code, src_dsm_dsuid, dest_dsm_dsuid)
{
}

dSMApiSelectiveListener::BusChangeData::BusChangeData(dsuid_t dsm_dsuid, int status) :
        CBData(dSMApiSelectiveListener::DsmApiBusChange),
        m_dsuid(dsm_dsuid), m_status(status)
{}

dsuid_t dSMApiSelectiveListener::BusChangeData::getDSUID() const
{
    return m_dsuid;
}

int dSMApiSelectiveListener::BusChangeData::getStatus() const
{
    return m_status;
}

bool dSMApiSelectiveListener::checkCB(
        dSMApiSelectiveListener::cb_selection sel)
{
    boost::recursive_mutex::scoped_lock lock(m_sel_access_mutex);
    if (m_selection[sel])
    {
        return true;
    }

    return false;
}

dSMApiSelectiveListener::dSMApiSelectiveListener()
{
    m_selection.resize(CB_MAX);
    for (size_t i = 0; i < m_selection.size(); i++)
    {
        m_selection[i] = false;
    }
}

void dSMApiSelectiveListener::select(
        dSMApiSelectiveListener::cb_selection cb, bool on)
{
    boost::recursive_mutex::scoped_lock lock(m_sel_access_mutex);
    m_selection[cb] = on;
}

dSMApiSelectiveListener::~dSMApiSelectiveListener() {}

void dSMApiSelectiveListener::linkStateCallback(bool _link)
{
    (void)_link;
    if (!checkCB(DsmApiLinkState))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::busChangeCallback(dsuid_t *id, int flag)
{
    if (!checkCB(DsmApiBusChange))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::BusChangeData> d(
                new dSMApiSelectiveListener::BusChangeData(*id, flag));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::busChangeCallback caught exception";
    }
}

void dSMApiSelectiveListener::deviceActiveCallback(uint8_t error_code,
                                    dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
                                    uint16_t device_id, uint16_t zone_id,
                                    uint16_t vendor_id, dsuid_t dsuid)
{
    if (!checkCB(DsmApiDeviceActive))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceActiveData> d(
                new dSMApiSelectiveListener::DeviceActiveData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            device_id, zone_id, vendor_id,
                                            dsuid));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::deviceAddedCallback caught exception";
    }
}

void dSMApiSelectiveListener::deviceInactiveCallback(uint8_t error_code,
                                        dsuid_t src_dsm_dsuid,
                                        dsuid_t dest_dsm_dsuid,
                                        uint16_t device_id,
                                        uint16_t zone_id, uint16_t vendor_id,
                                        dsuid_t dsuid)
{
    if (!checkCB(DsmApiDeviceInactive))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceInactiveData> d(
            new dSMApiSelectiveListener::DeviceInactiveData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            device_id, zone_id, vendor_id,
                                            dsuid));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::deviceRemovedCallback caught exception";
    }
}

void dSMApiSelectiveListener::devicePingCallback(uint8_t error_code,
                                        dsuid_t src_dsm_dsuid,
                                        dsuid_t dest_dsm_dsuid,
                                        uint16_t device_id,
                                        uint8_t present)
{
    if (!checkCB(DsmApiDevicePing))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DevicePingData> d(
            new dSMApiSelectiveListener::DevicePingData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            device_id, present));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR) <<
            "dSMApiSelectiveListener::devicePingCallback caught exception";
    }
}

void dSMApiSelectiveListener::deviceStatisticsCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t offset, uint16_t value)
{
    if (!checkCB(DsmApiDeviceStatistics))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceStatisticData> d(
            new dSMApiSelectiveListener::DeviceStatisticData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid, device_id,
                                            offset, value));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
}

void dSMApiSelectiveListener::deviceActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t group_id, uint8_t button_nr,
            uint8_t click_type)
{
    (void)error_code;
    (void)src_dsm_dsuid;
    (void)dest_dsm_dsuid;
    (void)device_id;
    (void)zone_id;
    (void)group_id;
    (void)button_nr;
    (void)click_type;
    if (!checkCB(DsmApiDeviceAction))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceLocalActionCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint16_t zone_id, uint8_t state)
{
    (void)error_code;
    (void)src_dsm_dsuid;
    (void)dest_dsm_dsuid;
    (void)device_id;
    (void)zone_id;
    (void)state;
    if (!checkCB(DsmApiDeviceLocalAction))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceConfigCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t config_class, uint8_t config_index, uint8_t value)
{
    (void)error_code;
    (void)src_dsm_dsuid;
    (void)dest_dsm_dsuid;
    (void)device_id;
    (void)config_class;
    (void)config_index;
    (void)value;
    if (!checkCB(DsmApiDeviceConfig))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceTestTransmissionQualityCallback(
            uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t downstream_quality, uint16_t upstream_quality)
{
    if (!checkCB(DsmApiDeviceTestTransmissionQuality))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DeviceTestQualityData> d(
            new dSMApiSelectiveListener::DeviceTestQualityData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid, device_id,
                                            downstream_quality, upstream_quality));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
}

void dSMApiSelectiveListener::deviceRemovedByDsuidCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, dsuid_t dsuid)
{
    (void)error_code;
    (void)src_dsm_dsuid;
    (void)dest_dsm_dsuid;
    (void)dsuid;
    if (!checkCB(DsmApiDeviceRemovedByDsuid))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::deviceRemovedByIdCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id)
{
    (void)error_code;
    (void)src_dsm_dsuid;
    (void)dest_dsm_dsuid;
    (void)device_id;
    if (!checkCB(DsmApiDeviceRemovedById))
    {
        return;
    }

    FILE_LOG(logWARNING) << "Callback selection not implemented";
}

void dSMApiSelectiveListener::commUpstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t is_resend, uint8_t is_sensor,
            uint8_t key_type, uint8_t click_type, uint8_t quality,
            uint8_t flags, uint8_t cflags)
{
    if (!checkCB(DsmApiCommUpstreamShort))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommUpstreamShortData> d(
            new dSMApiSelectiveListener::CommUpstreamShortData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            device_id, circuit_id, is_resend,
                                            is_sensor, key_type, click_type,
                                            quality, flags, cflags));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commUpstreamShortCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commUpstreamLongCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint16_t device_id,
            uint8_t circuit_id, uint8_t mainqualifier, uint8_t subqualifier,
            uint32_t data, uint8_t quality, uint8_t flags, uint8_t cflags)
{
    if (!checkCB(DsmApiCommUpstreamLong))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommUpstreamLongData> d(
            new dSMApiSelectiveListener::CommUpstreamLongData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            device_id, circuit_id,
                                            mainqualifier, subqualifier,
                                            data, quality, flags, cflags));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commUpstreamLongCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamShortCallback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
    if (!checkCB(DsmApiCommDownstreamShort))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamShortData> d(
            new dSMApiSelectiveListener::CommDownstreamShortData(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            command));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamShortCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamLong0Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command)
{
    if (!checkCB(DsmApiCommDownstreamLong0))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamLong0Data> d(
            new dSMApiSelectiveListener::CommDownstreamLong0Data(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            command));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamLong0Callback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamLong8Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint8_t data)
{
    if (!checkCB(DsmApiCommDownstreamLong8))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamLong8Data> d(
            new dSMApiSelectiveListener::CommDownstreamLong8Data(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            command, data));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamLong8Callback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::commDownstreamLong16Callback(uint8_t error_code,
            dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid, uint8_t command,
            uint16_t data)
{
    if (!checkCB(DsmApiCommDownstreamLong16))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::CommDownstreamLong16Data> d(
            new dSMApiSelectiveListener::CommDownstreamLong16Data(error_code,
                                            src_dsm_dsuid, dest_dsm_dsuid,
                                            command, data));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::commDownstreamLong16Callback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::dsmStatusCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint8_t hardware, uint8_t communicationdsp,
        uint8_t temperature, uint8_t configuration,
        uint8_t ds485bus, uint8_t powerline)
{
    if (!checkCB(DsmApiDsmState))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DsmStatusData> d(
            new dSMApiSelectiveListener::DsmStatusData(error_code,
                    src_dsm_dsuid, dest_dsm_dsuid,
                    hardware, communicationdsp, temperature,
                    configuration, ds485bus, powerline));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::dsmStatusCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::dsmErrorCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid,
        uint16_t error, uint32_t data)
{
    if (!checkCB(DsmApiDsmError))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DsmErrorData> d(
            new dSMApiSelectiveListener::DsmErrorData(error_code,
                    src_dsm_dsuid, dest_dsm_dsuid,
                    error, data));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::dsmErrorCallback"
            << " caught exception";
    }
}

void dSMApiSelectiveListener::dsmRebootRequestCallback(uint8_t error_code,
        dsuid_t src_dsm_dsuid, dsuid_t dest_dsm_dsuid)
{
    if (!checkCB(DsmApiDsmRebootRequest))
    {
        return;
    }

    try
    {
        boost::shared_ptr<dSMApiSelectiveListener::DsmRebootRequestData> d(
            new dSMApiSelectiveListener::DsmRebootRequestData(error_code,
                    src_dsm_dsuid, dest_dsm_dsuid));
        callback(boost::dynamic_pointer_cast<CBData>(d));
    }
    catch (Exception &ex)
    {
        FILE_LOG(logERROR) << ex.getMessage();
    }
    catch (std::exception &e)
    {
        FILE_LOG(logERROR) << e.what();
    }
    catch (...)
    {
        FILE_LOG(logERROR)
            << "dSMApiSelectiveListener::dsmRebootRequestCallback"
            << " caught exception";
    }
}

}; // namespace

