/*
    Copyright (c) 2015 digitalSTROM AG, Zurich, Switzerland

    This file is part of dS Assistant.

    dS-Assistant is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    dS-Assistant is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with digitalSTROM Server. If not, see <http://www.gnu.org/licenses/>.

    NOTE: all APIs, especially the JSON API, offered by this application
          are subject to change without further notice.
*/

#ifndef __TASKS_DESC__
#define __TASKS_DESC__

#include "timer.h"

namespace dSA
{

// forward declaration
class dS_Assistant;


#define INVALID_TASK_ID 0

// these values are saved in the database, so make sure to only add types,
// do not change or modify existing types
typedef enum
{
    UnknownTask             = 0,
    TestTransmissionQuality = 1,
    TestDevicePresence      = 2,
    TestDeviceStatistic     = 3,
    TestMeterStatistic      = 4,
    TestSystemMonitor       = 5,
    TestDeviceStatisticReset= 6,
    TestMeterStatisticReset = 7,
    TestOneDevicePresence   = 8,
    TestDeviceMonitor       = 9,
    TestDeviceUpdateTask    = 10,
    ModifyDevicePulse       = 11,
    ResetDevicePulse        = 12,
    ReadAreaConfiguration   = 13,
    TestFilterStatistic     = 14
} task_type;

typedef enum
{
    UnknownStatus       = 0,
    Queued              = 1,
    Running             = 2,
    Finished            = 3,
    Cancelled           = 4,
    Sleeping            = 5,
    Paused              = 6,
    Suspended           = 7
} task_status;

typedef enum
{
    Normal  = 0,
    High    = 1
} task_priority;

typedef enum
{
    kStateInit,
    kStateSetup,
    kStateStartup,
    kStateStartupWait,
    kStateStartupComplete,
    kStateCleanup,
    kStateCleanupWait,
    kStateVersion,
    kStateRequest,
    kStateResponse,
    kStateReady,
    kStateNext,
    kStateWait,
    kStateFinished
} task_state;

class dSTaskInfo
{
public:
    dSTaskInfo();
    void setDescription(std::string description);
    const std::string getName() const;
    const std::string getDescription() const;
    bool isValid() const;
    virtual void invalidate(bool shutdown = false);

    virtual ~dSTaskInfo();

protected:
    std::string m_name;
    std::string m_description;
    bool m_valid;
};

class dSLongTaskInfo : public dSTaskInfo
{
public:
    dSLongTaskInfo();
    virtual ~dSLongTaskInfo();

    void setTaskID(long long int task_id);
    long long int getTaskID() const;
    void setParentTaskID(long long int task_id);
    long long int getParentTaskID() const;
    unsigned char getProgress() const;
    std::string getProgressInfo() const;
    std::string getStatusInfo() const;
    std::string getParameters() const;
    void setTaskStatus(task_status status);
    task_status getTaskStatus() const;
    std::string getTaskStatusText() const;
    virtual task_type getTaskType() const = 0;
    time_t getStartedTime() const;
    void setStartedTime(time_t t);
    time_t getFinishedTime() const;
    void setFinishedTime(time_t t);
    time_t getQueuedTime() const;
    void setQueuedTime(time_t t);
    bool isPausable() const;
    bool isPaused() const;
    void pause();
    void setPriority(task_priority priority);
    task_priority getPriority() const;
    bool isPersistent() const;

protected:
    long long int m_task_id;
    long long int m_parent_task_id;
    unsigned char m_progress;
    std::string m_progress_info;
    std::string m_status_info;
    task_status m_task_status;
    mutable boost::mutex m_status_mutex;
    std::string m_parameters;
    time_t m_started;
    time_t m_finished;
    time_t m_queued;
    bool m_paused;
    bool m_pausable;
    task_priority m_task_priority;
    bool m_persistent;
};

class dSTask : public dSTaskInfo
{
public:
    dSTask();
    virtual ~dSTask() {}
    virtual void run(dS_Assistant *) = 0;
};

class dSLongTask : public dSLongTaskInfo
{
public:
    dSLongTask();
    virtual ~dSLongTask() {}
    virtual void run(dS_Assistant *) = 0;
};

class dSPersistentTask : public dSLongTask
{
public:
    dSPersistentTask();
    virtual ~dSPersistentTask() {}
    virtual void run(dS_Assistant *) = 0;
};

class TaskScheduler : public TimerSubscriber
{
public:
    TaskScheduler(boost::shared_ptr<dSTask> task, unsigned int delay);
    virtual ~TaskScheduler() {}
    boost::shared_ptr<dSTask> getTask();
    virtual void notify();
private:
    boost::shared_ptr<dSTask> m_task;
};

class LongTaskScheduler : public TimerSubscriber
{
public:
    LongTaskScheduler(boost::shared_ptr<dSLongTask> task, unsigned int delay);
    virtual ~LongTaskScheduler() {}
    boost::shared_ptr<dSLongTask> getTask();
    virtual void notify();
private:
    boost::shared_ptr<dSLongTask> m_task;
};


}

#endif
